//
// Cforall Version 1.0.0 Copyright (C) 2015 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// FixInit.h --
//
// Author           : Rob Schluntz
// Created On       : Wed Jan 13 16:29:30 2016
// Last Modified By : Rob Schluntz
// Last Modified On : Mon Apr 25 14:44:21 2016
// Update Count     : 30
//

#include <stack>
#include <list>
#include "RemoveInit.h"
#include "ResolvExpr/Resolver.h"
#include "ResolvExpr/typeops.h"
#include "SynTree/Declaration.h"
#include "SynTree/Type.h"
#include "SynTree/Expression.h"
#include "SynTree/Statement.h"
#include "SynTree/Initializer.h"
#include "SynTree/Mutator.h"
#include "SymTab/Indexer.h"
#include "GenPoly/PolyMutator.h"

bool ctordtorp = false;
#define PRINT( text ) if ( ctordtorp ) { text }

namespace InitTweak {
	namespace {
		const std::list<Label> noLabels;
		const std::list<Expression*> noDesignators;
	}

	class InsertImplicitCalls : public Mutator {
	public:
		/// wrap function application expressions as ImplicitCopyCtorExpr nodes
		/// so that it is easy to identify which function calls need their parameters
		/// to be copy constructed
		static void insert( std::list< Declaration * > & translationUnit );

		virtual Expression * mutate( ApplicationExpr * appExpr );
	};

	class ResolveCopyCtors : public SymTab::Indexer {
	public:
		/// generate temporary ObjectDecls for each argument and return value of each
		/// ImplicitCopyCtorExpr, generate/resolve copy construction expressions for each,
		/// and generate/resolve destructors for both arguments and return value temporaries
		static void resolveImplicitCalls( std::list< Declaration * > & translationUnit );

		virtual void visit( ImplicitCopyCtorExpr * impCpCtorExpr );

		/// create and resolve ctor/dtor expression: fname(var, [cpArg])
		ApplicationExpr * makeCtorDtor( const std::string & fname, ObjectDecl * var, Expression * cpArg = NULL );
	};

	class FixInit : public GenPoly::PolyMutator {
	  public:
		/// expand each object declaration to use its constructor after it is declared.
		/// insert destructor calls at the appropriate places
		static void fixInitializers( std::list< Declaration * > &translationUnit );

		virtual DeclarationWithType * mutate( ObjectDecl *objDecl );

		virtual CompoundStmt * mutate( CompoundStmt * compoundStmt );
		virtual Statement * mutate( ReturnStmt * returnStmt );
		virtual Statement * mutate( BranchStmt * branchStmt );

	  private:
		// stack of list of statements - used to differentiate scopes
		std::list< std::list< Statement * > > dtorStmts;
	};

	class FixCopyCtors : public GenPoly::PolyMutator {
	  public:
		/// expand ImplicitCopyCtorExpr nodes into the temporary declarations, copy constructors,
		/// call expression, and destructors
		static void fixCopyCtors( std::list< Declaration * > &translationUnit );

		virtual Expression * mutate( ImplicitCopyCtorExpr * impCpCtorExpr );

	  private:
		// stack of list of statements - used to differentiate scopes
		std::list< std::list< Statement * > > dtorStmts;
	};

	void fix( std::list< Declaration * > & translationUnit ) {
		InsertImplicitCalls::insert( translationUnit );
		ResolveCopyCtors::resolveImplicitCalls( translationUnit );
		FixInit::fixInitializers( translationUnit );
		// FixCopyCtors must happen after FixInit, so that destructors are placed correctly
		FixCopyCtors::fixCopyCtors( translationUnit );
	}

	void InsertImplicitCalls::insert( std::list< Declaration * > & translationUnit ) {
		InsertImplicitCalls inserter;
		mutateAll( translationUnit, inserter );
	}

	void ResolveCopyCtors::resolveImplicitCalls( std::list< Declaration * > & translationUnit ) {
		ResolveCopyCtors resolver;
		acceptAll( translationUnit, resolver );
	}

	void FixInit::fixInitializers( std::list< Declaration * > & translationUnit ) {
		FixInit fixer;
		mutateAll( translationUnit, fixer );
	}

	void FixCopyCtors::fixCopyCtors( std::list< Declaration * > & translationUnit ) {
		FixCopyCtors fixer;
		mutateAll( translationUnit, fixer );
	}

	Expression * InsertImplicitCalls::mutate( ApplicationExpr * appExpr ) {
		appExpr = dynamic_cast< ApplicationExpr * >( Mutator::mutate( appExpr ) );
		assert( appExpr );

		if ( VariableExpr * function = dynamic_cast< VariableExpr * > ( appExpr->get_function() ) ) {
			if ( function->get_var()->get_linkage() == LinkageSpec::Intrinsic ) {
				// optimization: don't need to copy construct in order to call intrinsic functions
				return appExpr;
			} else if ( FunctionDecl * funcDecl = dynamic_cast< FunctionDecl * > ( function->get_var() ) ) {
				FunctionType * ftype = funcDecl->get_functionType();
				if ( (funcDecl->get_name() == "?{}" || funcDecl->get_name() == "?=?") && ftype->get_parameters().size() == 2 ) {
					Type * t1 = ftype->get_parameters().front()->get_type();
					Type * t2 = ftype->get_parameters().back()->get_type();
					PointerType * ptrType = dynamic_cast< PointerType * > ( t1 );
					assert( ptrType );
					if ( ResolvExpr::typesCompatible( ptrType->get_base(), t2, SymTab::Indexer() ) ) {
						// optimization: don't need to copy construct in order to call a copy constructor or
						// assignment operator
						return appExpr;
					}
				} else if ( funcDecl->get_name() == "^?{}" ) {
					// correctness: never copy construct arguments to a destructor
					return appExpr;
				}
			}
		}
		PRINT( std::cerr << "InsertImplicitCalls: adding a wrapper " << appExpr << std::endl; )

		// wrap each function call so that it is easy to identify nodes that have to be copy constructed
		return new ImplicitCopyCtorExpr( appExpr );
	}

	ApplicationExpr * ResolveCopyCtors::makeCtorDtor( const std::string & fname, ObjectDecl * var, Expression * cpArg ) {
		assert( var );
		UntypedExpr * untyped = new UntypedExpr( new NameExpr( fname ) );
		untyped->get_args().push_back( new AddressExpr( new VariableExpr( var ) ) );
		if (cpArg) untyped->get_args().push_back( cpArg );

		// resolve copy constructor
		// should only be one alternative for copy ctor and dtor expressions, since
		// all arguments are fixed (VariableExpr and already resolved expression)
		PRINT( std::cerr << "ResolvingCtorDtor " << untyped << std::endl; )
		ApplicationExpr * resolved = dynamic_cast< ApplicationExpr * >( ResolvExpr::findVoidExpression( untyped, *this ) );

		assert( resolved );
		delete untyped;
		return resolved;
	}

	void ResolveCopyCtors::visit( ImplicitCopyCtorExpr *impCpCtorExpr ) {
		static UniqueName tempNamer("_tmp_cp");
		static UniqueName retNamer("_tmp_cp_ret");

		PRINT( std::cerr << "ResolveCopyCtors: " << impCpCtorExpr << std::endl; )
		Visitor::visit( impCpCtorExpr );

		ApplicationExpr * appExpr = impCpCtorExpr->get_callExpr();

		// take each argument and attempt to copy construct it.
		for ( Expression * & arg : appExpr->get_args() ) {
			// xxx - need to handle tuple arguments
			assert( ! arg->get_results().empty() );
			ObjectDecl * tmp = new ObjectDecl( tempNamer.newName(), DeclarationNode::NoStorageClass, LinkageSpec::C, 0, arg->get_results().front()->clone(), 0 );
			tmp->get_type()->set_isConst( false );

			// create and resolve copy constructor
			PRINT( std::cerr << "makeCtorDtor for an argument" << std::endl; )
			ApplicationExpr * cpCtor = makeCtorDtor( "?{}", tmp, arg );

			// if the chosen constructor is intrinsic, the copy is unnecessary, so
			// don't create the temporary and don't call the copy constructor
			VariableExpr * function = dynamic_cast< VariableExpr * >( cpCtor->get_function() );
			assert( function );
			if ( function->get_var()->get_linkage() != LinkageSpec::Intrinsic ) {
				// replace argument to function call with temporary
				arg = new VariableExpr( tmp );
				impCpCtorExpr->get_tempDecls().push_back( tmp );
				impCpCtorExpr->get_copyCtors().push_back( cpCtor );
				impCpCtorExpr->get_dtors().push_front( makeCtorDtor( "^?{}", tmp ) );
			}
		}

		// each return value from the call needs to be connected with an ObjectDecl
		// at the call site, which is initialized with the return value and is destructed
		// later
		// xxx - handle multiple return values
		ApplicationExpr * callExpr = impCpCtorExpr->get_callExpr();
		for ( Type * result : appExpr->get_results() ) {
			ObjectDecl * ret = new ObjectDecl( retNamer.newName(), DeclarationNode::NoStorageClass, LinkageSpec::C, 0, result->clone(), new SingleInit( callExpr ) );
			ret->get_type()->set_isConst( false );
			impCpCtorExpr->get_returnDecls().push_back( ret );
			PRINT( std::cerr << "makeCtorDtor for a return" << std::endl; )
			impCpCtorExpr->get_dtors().push_front( makeCtorDtor( "^?{}", ret ) );
		}
		PRINT( std::cerr << "after Resolving: " << impCpCtorExpr << std::endl; )
	}


	Expression * FixCopyCtors::mutate( ImplicitCopyCtorExpr * impCpCtorExpr ) {
		PRINT( std::cerr << "FixCopyCtors: " << impCpCtorExpr << std::endl; )

		// assert( impCpCtorExpr->get_callExpr()->get_env() );
		impCpCtorExpr = dynamic_cast< ImplicitCopyCtorExpr * >( Mutator::mutate( impCpCtorExpr ) );
		assert( impCpCtorExpr );

		std::list< Expression * > & copyCtors = impCpCtorExpr->get_copyCtors();
		std::list< ObjectDecl * > & tempDecls = impCpCtorExpr->get_tempDecls();
		std::list< ObjectDecl * > & returnDecls = impCpCtorExpr->get_returnDecls();
		std::list< Expression * > & dtors = impCpCtorExpr->get_dtors();

		// add all temporary declarations and their constructors
		for ( ObjectDecl * obj : tempDecls ) {
			assert( ! copyCtors.empty() );
			stmtsToAdd.push_back( new DeclStmt( noLabels, obj ) );
			stmtsToAdd.push_back( new ExprStmt( noLabels, copyCtors.front() ) );
			copyCtors.pop_front();
		}

		// add destructors after current statement
		for ( Expression * dtor : dtors ) {
			stmtsToAddAfter.push_back( new ExprStmt( noLabels, dtor ) );
		}

		// xxx - update to work with multiple return values
		ObjectDecl * returnDecl = returnDecls.empty() ? NULL : returnDecls.front();
		Expression * callExpr = impCpCtorExpr->get_callExpr();

		PRINT( std::cerr << "Coming out the back..." << impCpCtorExpr << std::endl; )

		// xxx - some of these aren't necessary, and can be removed once this is stable
		copyCtors.clear();
		dtors.clear();
		tempDecls.clear();
		returnDecls.clear();
		impCpCtorExpr->set_callExpr( NULL );
		delete impCpCtorExpr;

		if ( returnDecl ) {
			// call is currently attached to first returnDecl
			stmtsToAdd.push_back( new DeclStmt( noLabels, returnDecl ) );
			return new VariableExpr( returnDecl );
		} else {
			// add call expression - if no return values, can call directly
			assert( callExpr );
			return callExpr;
		}
	}

	DeclarationWithType *FixInit::mutate( ObjectDecl *objDecl ) {
		// first recursively handle pieces of ObjectDecl so that they aren't missed by other visitors
		// when the init is removed from the ObjectDecl
		objDecl = dynamic_cast< ObjectDecl * >( Mutator::mutate( objDecl ) );

		if ( ConstructorInit * ctorInit = dynamic_cast< ConstructorInit * >( objDecl->get_init() ) ) {
			// a decision should have been made by the resolver, so ctor and init are not both non-NULL
			assert( ! ctorInit->get_ctor() || ! ctorInit->get_init() );
			if ( Statement * ctor = ctorInit->get_ctor() ) {
				if ( objDecl->get_storageClass() == DeclarationNode::Static ) {
					// generate:
					// static bool __objName_uninitialized = true;
					// if (__objName_uninitialized) {
					//   __ctor(__objName);
					//   void dtor_atexit() {
					//     __dtor(__objName);
					//   }
					//   on_exit(dtorOnExit, &__objName);
					//   __objName_uninitialized = false;
					// }

					// generate first line
					BasicType * boolType = new BasicType( Type::Qualifiers(), BasicType::Bool );
					SingleInit * boolInitExpr = new SingleInit( new ConstantExpr( Constant( boolType->clone(), "1" ) ), noDesignators );
					ObjectDecl * isUninitializedVar = new ObjectDecl( objDecl->get_mangleName() + "_uninitialized", DeclarationNode::Static, LinkageSpec::Cforall, 0, boolType, boolInitExpr );
					isUninitializedVar->fixUniqueId();

					// void dtor_atexit(...) {...}
					FunctionDecl * dtorCaller = new FunctionDecl( objDecl->get_mangleName() + "_dtor_atexit", DeclarationNode::NoStorageClass, LinkageSpec::C, new FunctionType( Type::Qualifiers(), false ), new CompoundStmt( noLabels ), false, false );
					dtorCaller->fixUniqueId();
					dtorCaller->get_statements()->get_kids().push_back( ctorInit->get_dtor() );

					// on_exit(dtor_atexit);
					UntypedExpr * callAtexit = new UntypedExpr( new NameExpr( "atexit" ) );
					callAtexit->get_args().push_back( new VariableExpr( dtorCaller ) );

					// __objName_uninitialized = false;
					UntypedExpr * setTrue = new UntypedExpr( new NameExpr( "?=?" ) );
					setTrue->get_args().push_back( new VariableExpr( isUninitializedVar ) );
					setTrue->get_args().push_back( new ConstantExpr( Constant( boolType->clone(), "0" ) ) );

					// generate body of if
					CompoundStmt * initStmts = new CompoundStmt( noLabels );
					std::list< Statement * > & body = initStmts->get_kids();
					body.push_back( ctor );
					body.push_back( new DeclStmt( noLabels, dtorCaller ) );
					body.push_back( new ExprStmt( noLabels, callAtexit ) );
					body.push_back( new ExprStmt( noLabels, setTrue ) );

					// put it all together
					IfStmt * ifStmt = new IfStmt( noLabels, new VariableExpr( isUninitializedVar ), initStmts, 0 );
					stmtsToAddAfter.push_back( new DeclStmt( noLabels, isUninitializedVar ) );
					stmtsToAddAfter.push_back( ifStmt );
				} else {
					stmtsToAddAfter.push_back( ctor );
					dtorStmts.back().push_front( ctorInit->get_dtor() );
				}
				objDecl->set_init( NULL );
				ctorInit->set_ctor( NULL );
				ctorInit->set_dtor( NULL );  // xxx - only destruct when constructing? Probably not?
			} else if ( Initializer * init = ctorInit->get_init() ) {
				objDecl->set_init( init );
				ctorInit->set_init( NULL );
			} else {
				// no constructor and no initializer, which is okay
				objDecl->set_init( NULL );
			}
			delete ctorInit;
		}
		return objDecl;
	}

	template<typename Iterator, typename OutputIterator>
	void insertDtors( Iterator begin, Iterator end, OutputIterator out ) {
		for ( Iterator it = begin ; it != end ; ++it ) {
			// remove if instrinsic destructor statement
			// xxx - test user manually calling intrinsic functions - what happens?
			if ( ExprStmt * exprStmt = dynamic_cast< ExprStmt * >( *it ) ) {
				ApplicationExpr * appExpr = dynamic_cast< ApplicationExpr * >( exprStmt->get_expr() );
				assert( appExpr );
				VariableExpr * function = dynamic_cast< VariableExpr * >( appExpr->get_function() );
				assert( function );
				// check for Intrinsic only - don't want to remove all overridable dtors because autogenerated dtor
				// will call all member dtors, and some members may have a user defined dtor.
				if ( function->get_var()->get_linkage() == LinkageSpec::Intrinsic ) {
					// don't need to call intrinsic dtor, because it does nothing
				} else {
					// non-intrinsic dtors must be called
					*out++ = (*it)->clone();
				}
			} else {
				// could also be a compound statement with a loop, in the case of an array
				*out++ = (*it)->clone();
			}
		}
	}


	CompoundStmt * FixInit::mutate( CompoundStmt * compoundStmt ) {
		// mutate statements - this will also populate dtorStmts list.
		// don't want to dump all destructors when block is left,
		// just the destructors associated with variables defined in this block,
		// so push a new list to the top of the stack so that we can differentiate scopes
		dtorStmts.push_back( std::list<Statement *>() );

		compoundStmt = PolyMutator::mutate( compoundStmt );
		std::list< Statement * > & statements = compoundStmt->get_kids();

		insertDtors( dtorStmts.back().begin(), dtorStmts.back().end(), back_inserter( statements ) );

		deleteAll( dtorStmts.back() );
		dtorStmts.pop_back();
		return compoundStmt;
	}

	Statement * FixInit::mutate( ReturnStmt * returnStmt ) {
		for ( std::list< std::list< Statement * > >::reverse_iterator list = dtorStmts.rbegin(); list != dtorStmts.rend(); ++list ) {
			insertDtors( list->begin(), list->end(), back_inserter( stmtsToAdd ) );
		}
		return Mutator::mutate( returnStmt );
	}

	Statement * FixInit::mutate( BranchStmt * branchStmt ) {
		// TODO: adding to the end of a block isn't sufficient, since
		// return/break/goto should trigger destructor when block is left.
		switch( branchStmt->get_type() ) {
			case BranchStmt::Continue:
			case BranchStmt::Break:
				insertDtors( dtorStmts.back().begin(), dtorStmts.back().end(), back_inserter( stmtsToAdd ) );
				break;
			case BranchStmt::Goto:
				// xxx
				// if goto leaves a block, generate dtors for every block it leaves
				// if goto is in same block but earlier statement, destruct every object that was defined after the statement
				break;
			default:
				assert( false );
		}
		return Mutator::mutate( branchStmt );
	}


} // namespace InitTweak

// Local Variables: //
// tab-width: 4 //
// mode: c++ //
// compile-command: "make install" //
// End: //
